#!/usr/bin/env python

from datetime import datetime, timedelta
import os

from scipy.interpolate import SmoothBivariateSpline
from xarray import DataArray, Dataset

import matplotlib.pyplot as plt
import numpy as np
import scipy.integrate
import scipy.ndimage as ndi
import skimage.morphology
import xarray

import doas
import dplume

import c2sm.misc.conv

from smartcarb import config as cfg
from smartcarb import io
from smartcarb import misc
from smartcarb import map


import scipy.optimize as sopt



# xlim and ylim for visualization of Gauss plume
FILENAME2LIM = {
    'Sentinel_7_CO2_2015011311_o0190_l270.nc': (-15,90,-15,90, 20),
    'Sentinel_7_CO2_2015012411_o0354_l270.nc': (-25,25,-15,110, 20),
}



class DataVector:
    def __init__(self, x, y, a):
        self.x = x
        self.y = y
        self.a = a


def distance(q,p):
    """Distance between two points."""
    return np.sqrt((q[0] - p[0])**2 + (q[1] - p[1])**2)


def find_min_distance_to_curve(p, curve):
    """
    p :: point x0,y0
    f :: function
    c :: coefficients of 2D curve
    """
    s = 1e-3
    constr = [{
        'type': 'eq',
        'fun':  lambda q: s * (curve(x=q[0], y=q[1])[0] - q[0]),
    },{
        'type': 'eq',
        'fun':  lambda q: s * (curve(x=q[0], y=q[1])[1] - q[1]),
    }]

    q0 = curve(x=p[0], y=p[1])

    res = sopt.minimize(distance, x0=q0, args=(p,),
            method='SLSQP', constraints=constr[0],
            options={'maxiter': 100},
    )

    if not res.success:
        print(res)

    return res.x, res.fun, res


def kgs_to_Mtyr(x, inverse=False):
    """ Convert kg/s to Mt/yr. """
    factor = 1e9 / c2sm.misc.conv.SECONDS_PER_YEAR
    if inverse:
        return x * factor
    else:
        return x / factor


def gauss1d(Q,u):
    """
    Returns Gauss plume integrated in y- and z- direction along x-axis.

    Q: emissions in kg/s
    u: wind speed in m/s
    """
    return Q / u



def gauss2d(p, x, y, u, x0=0.0, y0=0.0):
    """\
    Computes Gaussian plume (units: kg/m2).

    p: state vector (Q, K, BG)
    - Q:  emissions in kg/s
    - K:  eddy diffusion coefficient in m2/s
    - BG: background field in kg/m2

    x: distance from origin in m
    y: distance from center line in m
    u: wind speed in m/s

    x0: x-coordinate of origin
    y0: y-coordinate of origin

    TODO:
    - decay time

    """
    # state vector
    Q = p[0]        # emissions (kg/s)
    K = p[1]        # eddy diffusion coefficient (m2 / s)
    BG = p[2]       # background field (kg/m2)

    # dispersion along the plume
    sigma = np.sqrt(2.0 * K * (x-x0) / u)

    # compute Gaussian plume
    A = Q / (np.sqrt(2.0 * np.pi) * sigma * u) * np.exp(-0.5 * (y-y0)**2 / sigma**2)

    plume = np.full(x.shape, BG)
    plume[x > 0.0] += A[x > 0.0]

    return plume



def find_all_roots(hh, dd, d, h_max=5000.0):
    """\
    Find all heights (hh), below hh_max where wind direction (dd) equals
    given wind direction (d).
    """

    founds = []

    for i in range(hh.size-1):
        if dd[i] < dd[i+1]:
            dmin, dmax = dd[i], dd[i+1]
            hmin, hmax = hh[i], hh[i+1]
        else:
            dmin, dmax = dd[i+1], dd[i]
            hmin, hmax = hh[i+1], hh[i]

        if hmin > h_max:
            continue

        if dmin <= d < dmax:
            d0 = np.interp(d, [dmin, dmax], [hmin, hmax])
            founds.append(d0)

    return founds




def arc_length_of_poly(f, c, a, b):

    x = np.linspace(a,b)

    s = np.sqrt( 1 + np.polyval(np.polyder(c), x)**2 )

    return scipy.integrate.simps(s,x)


def arc_length_of_2d_curve(curve, a, b):

    t = np.linspace(a,b, 5001)
    xt, yt = curve(t=t, m=1)

    v = np.sqrt( xt**2 + yt**2)

    l = scipy.integrate.simps(v,t)
    return l




def compute_arc_length_and_distance(points, curve):
    """\
    Given n points and polynomial function on [x_o, x_max], compute
    arc length along curve and distance from curve.
    """
    points = points

    dxs = []
    dys = []

    t_o = curve.get_parameter(curve.x_o, curve.y_o)

    for px,py in points:
        (qx,qy), dy, res = find_min_distance_to_curve([px,py], curve)

        t = curve.get_parameter(qx, qy)
        dx = arc_length_of_2d_curve(curve, t_o, t)

        dxs.append(dx)
        dys.append(dy)

    return np.array(dxs), np.array(dys)




def extract_along_curve(x,y, field):
    pass


def flux_inversion(p0, dx, dy, field, wind, std=0.7):

    def model(x,b):
        return gauss2d(x,**b)

    p0 = np.asarray(p0)
    Se = np.full(field.shape, std**2)

    b = {'x': dx, 'y': dy, 'u': wind}
    x, res = doas.solver.gauss_newton(p0, field, model, Se=Se, b=b, Kb_keys=['u'])


    G = doas.solver.gain_matrix(res['K'], Se)


    #X2 = []
    #for i in range(100):
    #    b = {'x': dx, 'y': dy, 'u': wind + np.random.randn()}
    #    x2, _ = doas.solver.gauss_newton(p0, field, model, Se=Se, b=b, Kb_keys=['u'])
    #    X2.append(x2)
    #print('sigma_b =', np.array(X2).std(0))


    return x, res



def compute_2D_curve(lon_or_x, lat_or_y, weights, lon_o=None, lat_o=None,
        origin=None, degree=2, epsg=cfg.EPSG):
    """\
    Fit a 2D Poly.
    """
    if epsg is None:
        x, y = lon_or_x, lat_or_y
    else:
        if origin is not None:
            lon_o, lat_o = misc.get_source_origin(origin)

        x, y = misc.xy2wgs(lon_or_x, lat_or_y, origin, epsg, inverse=True)

    x0 = -np.sum(x*weights) / np.sum(weights)
    y0 = -np.sum(y*weights) / np.sum(weights)
    curve = Poly2D(x, y, weights, degree, 0.0, 0.0, x0, y0)

    t = curve.get_parameter(x, y)
    t = np.linspace(t.min(), t.max())
    x_curve, y_curve = curve(t=t)

    return x_curve, y_curve, curve



class Poly2D:
    def __init__(self, x, y, w, degree=2, x_o=0.0, y_o=0.0,
            x0=0.0, y0=0.0):
        """\
        A 2D curve fitted on the point cloud given by x and y.

        Parameters
        ----------
        x,y,w: x,y coords and weights used for fitting the data

        degree: degrees if the two polynomials x(t) and y(t)

        x_o, y_o: location of source (will be added to x,y and given
                  high weight such that curves goes through source)

        x0, y0: origin of coordinate system
        """
        self.x = x
        self.y = y
        self.w = w
        self.degree = degree

        self.x_o = x_o
        self.y_o = y_o
        self.t_o = np.nan

        self.x0 = x0
        self.y0 = y0

        self.c = np.zeros(2 * (degree + 1))

        self._fit()

        # arc length to origin
        self.t_o = self.get_parameter(self.x_o, self.y_o)

        # 
        self.tmin = 0.0
        self.tmax = np.max(self.get_parameter(self.x, self.y))
        self.interpolator = None


    def _fit(self):

        def objective(c, w, x, y):
            xt,yt = self._compute_curve(c,x,y)
            return np.concatenate([
                w * (xt - x), w * (yt - y)
            ])

        # add origin
        x = np.append(self.x, self.x_o)
        y = np.append(self.y, self.y_o)
        w = np.append(self.w, 1000.)

        # angle around origin (not used)
        #phi = np.arctan2(y - y0, x - x0)

        # curve fit
        self.c, res = sopt.leastsq(objective, x0=self.c, args=(w,x,y))

        if res not in [1,2,3,4]:
            print('least square failed with error code: ', res)


    def arc2parameter(self, arc):

        if self.interpolator is None:
            ta = np.arange(self.tmin, 2 * self.tmax)
            la = [arc_length_of_2d_curve(self, self.t_o, v) for v in ta]
            self.interpolator = scipy.interpolate.interp1d(la, ta)

        return self.interpolator(arc)


    def get_parameter(self, x, y):
        return np.sqrt((x - self.x0)**2 + (y - self.y0)**2)


    def compute_normal(self, t0, x=None, y=None, t=None):

        x0, y0 = self(t=t0)

        # tangent vector
        v = np.array(self(t=t0, m=1))
        v /= np.linalg.norm(v)

        # rotate 90 degree
        n = np.dot([[0,-1],[1,0]], v)

        cx = np.array([n[0], x0])
        cy = np.array([n[1], y0])

        if t is None:
            s = np.sign(v[0] * (y - y0) - v[1] * (x - x0))
            t = s * np.sqrt((x-x0)**2 + (y-y0)**2)

        xt = np.polyval(cx,t)
        yt = np.polyval(cy,t)

        return xt,yt


    def _compute_curve(self, c, x=None, y=None, t=None, m=0):

        if t is None:
            t = self.get_parameter(x,y)
        else:
            if x is not None or y is not None:
                print('Warning: `x` and `y` will be ignored as `t` was given.')

        k = c.size // 3
        cx = c[:k]
        cy = c[k:]

        if m != 0:
            cx = np.polyder(cx, m)
            cy = np.polyder(cy, m)

        xt = np.polyval(cx,t)
        yt = np.polyval(cy,t)

        return xt, yt

    def __call__(self, x=None, y=None, t=None, m=0):
        return self._compute_curve(self.c, x, y, t, m)




def create_center_line(field, weights, degree, is_plume):
    # get location of plume and XCO2 values (plus wind etc.)
    x = np.asarray(field.x)[is_plume]
    y = np.asarray(field.y)[is_plume]
    w = np.asarray(weights)[is_plume]

    # origin of center line
    x0, y0 = -x.mean(), -y.mean()

    # fit 2d curve
    curve = Poly2D(x, y, w, degree, x_o=0.0, y_o=0.0, x0=x0, y0=y0)

    return curve


def setup_flux_inversion(data, origin, lon_model=None, lat_model=None,
                         height_model=None, uwind=None, vwind=None,
                         lon_o=None, lat_o=None,
                         degree=2, epsg=cfg.EPSG,
                         **kwargs):
    """
    LEVEL-2 data
    ------------
    data       :: level-2 data (xarray.Dataset):
    - co2      :: satellite observations (e.g. XCO2)
    - clouds   :: cloud fractions at ground pixels
    - is_plume :: boolean array where plume has been detected
    - z_values :: z-values from plume detection

    MODEL DATA (e.g COSMO)
    ----------------------
    lon_model    :: longitude in model (lon, lat)
    lat_model    :: latitude in model (lon, lat)
    height_model :: height in model (lon, lat, layers+1)
    uwind        :: u-wind in model (lon, lat, layers)
    vwind        :: v-wind in model (lon, lat, layers)

    FLUX INVERSION PARAMETERS
    -------------------------
    origin       :: name of plume (used to get lon_o and lat_o)
    lon_o, lat_o :: longitude/latitude of point source
    degree       :: degree of polynomial fitted on plume

    """

    if lon_model is None or lat_model is None:
        x_model, y_model = None, None
    else:
        x_model, y_model = misc.xy2wgs(lon_model, lat_model, origin, epsg, inverse=True)


    data['x'], data['y'] = misc.xy2wgs(data.lon, data.lat, origin,
                                       epsg=cfg.EPSG, inverse=True)


    # convert ppm to kg/m2
    data['CO2_mass'] = c2sm.misc.conv.ppm_to_mass(data.CO2, data.psurf, 'CO2')
    data['CO2_mass'].attrs['units'] = 'kg m-2'

    data['CO2_noise_mass'] = c2sm.misc.conv.ppm_to_mass(data.CO2_noise, data.psurf, 'CO2')
    data['CO2_noise_mass'].attrs['units'] = 'kg m-2'


    # DETERMINE CENTER LINE OF PLUME 
    weights = data.z_values #np.ones_like(data.z_values)
    curve = create_center_line(data, weights, degree, data.is_plume)

    # DISTANCE FROM CENTER LINE
    # make plume bigger to include background
    data['plume_area'] = xarray.full_like(data.is_plume, False)
    data['plume_area'][:] = skimage.morphology.dilation(data.is_plume, np.ones((10,10)))

    # compute for all ground pixels distance to center line
    x_a = np.array(data.x)[data.plume_area]
    y_a = np.array(data.y)[data.plume_area]
    points = np.transpose([x_a, y_a])

    xp = np.full(data.x.shape, np.nan)
    yp = np.full(data.x.shape, np.nan)

    xp[data.plume_area], yp[data.plume_area] = compute_arc_length_and_distance(points, curve)

    data['xp'] = xarray.full_like(data.x, np.nan)
    data['yp'] = xarray.full_like(data.y, np.nan)
    data.xp[:] = xp
    data.yp[:] = yp
    data.xp.attrs['long name'] = 'along plume coordiate'
    data.yp.attrs['long name'] = 'across plume coordinate'


    ###################################################
    # WIND SPEED ALONG CENTER LINE (TODO)             #
    ###################################################
    x_o, y_o = 0.0, 0.0

    if x_model is None:
        uu = uwind.copy()
        vv = vwind.copy()
        hh = height_model.copy()

        if hh[0] > hh[1]:
            hh = hh[::-1]
            uu = uu[::-1]
            vv = vv[::-1]

    else:
        i_o, j_o = misc.nearest_index(x_o, y_o, x_model, y_model)

        # wind profile at point source (TODO: spatial interpolation)
        uu = uwind[::-1, i_o, j_o]
        vv = vwind[::-1, i_o, j_o]
        hh = height_model[::-1, i_o, j_o]


    hh = 0.5 * (hh[1:] + hh[:-1])
    dd = (270.0 - np.rad2deg(np.arctan2(vv,uu))) % 360

    # wind direction (satellite)
    xt,yt = curve(x_o, y_o, m=1)
    d_sat = (270 - np.rad2deg(np.arctan2(yt,xt))) % 360

    # mean wind
    wind = np.sqrt((uu[hh <= 2000])**2 + np.mean(vv[hh <= 2000])**2)
    print('mean wind:', wind.mean(), wind.std())

    # wind from satellite wind direction
    wind = estimate_wind_speed(hh, uu, vv, d_sat)

    # TODO: compare estimated height with emission height in emission files
    plume_age = data.xp.max() / wind / 3600
    print('plume length:\t%d km' % (1e-3 * data.xp.max()))
    print('plume age:\t%.1f hours' % plume_age)

    params = {
        'wind': wind,
        'direction': d_sat,
    }


    return data, curve, params




def estimate_wind_speed(hh,uu,vv,d_sat, hmin=0, hmax=2000.0):

    dd = (270.0 - np.rad2deg(np.arctan2(vv,uu))) % 360

    # brute-force match wind direction between plume and model
    k_o = np.argmin(
            misc.anglediff(dd[(hmin <= hh) & (hh < hmax)], d_sat)**2
    )
    wind = np.sqrt(uu[k_o]**2 + vv[k_o]**2)

    print('first estimate:', hh[k_o], wind)

    # fine-tuning
    hh2 = np.linspace(hh[k_o-1], hh[k_o+1], 51)
    uu2 = np.interp(hh2, hh, uu)
    vv2 = np.interp(hh2, hh, vv)
    dd2 = (270.0 - np.rad2deg(np.arctan2(vv2,uu2))) % 360

    k2 = np.argmin(misc.anglediff(dd2, d_sat)**2)

    wind = np.sqrt(uu2[k2]**2 + vv2[k2]**2)

    print('fine estimate:', hh2[k2], wind)

    return wind


def draw_situation(field_vector, plume_vector, curve, ax=None, ms=20,
                   vmin=None, vmax=None):

    t = curve.get_parameter(field_vector.x, field_vector.y)
    t = np.linspace(t.min(), t.max())

    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111, aspect='equal')
    else:
        fig = plt.gcf()

    ax.plot(*curve(t=t), 'k-')

    c = ax.scatter(field_vector.x, field_vector.y, c=field_vector.a,
                   vmin=vmin, vmax=vmax, s=ms)

    ax.scatter(plume_vector.x, plume_vector.y, edgecolors='#222222',
               c='none', s=ms)

    ax.grid(True)
    ax.set_xlabel('x (km)')
    ax.set_ylabel('y (km)')

    fig.subplots_adjust(right=0.8)
    cax = fig.add_axes([0.82, 0.15, 0.03, 0.7])
    cb = fig.colorbar(c, cax=cax, extend='both')
    cb.set_label('XCO$_2$ (ppm)')



def visualize(x_c, y_c, field, x, dx, dy, wind, plume, filename):

    v = round(np.mean(field.a))
    s = max(1, round(np.std(field.a)))
    vmin, vmax = v-s, v+s

    fig1 = plt.figure()
    fig2 = plt.figure()

    ax1 = fig1.add_subplot(111, aspect='equal')
    ax2 = fig2.add_subplot(111, aspect='equal')


    ax1.set_ylabel('y (km)')
    ax1.set_title('(a) Observations')
    ax2.set_title('(b) Fit')

    try:
        xmin,xmax,ymin,ymax,ms = FILENAME2LIM[filename]
    except KeyError:
        xmin=xmax=ymin=ymax=None
        ms = None

    c = ax1.scatter(field.x, field.y, c=field.a, vmin=vmin, vmax=vmax, s=ms)
    c = ax2.scatter(field.x, field.y, c=gauss2d(x,dx,dy, u=wind),
                    vmin=vmin, vmax=vmax, s=ms)
    ax1.scatter(plume.x, plume.y, edgecolors='#222222', c='none', s=ms)


    for ax in [ax1, ax2]:

        ax.set_xlim(xmin,xmax)
        ax.set_ylim(ymin,ymax)

        ax.set_xlabel('x (km)')
        ax.plot(x_c, y_c, 'k-')

        # draw arrow
        d1,d2 = x_c[-1] - x_c[-2], y_c[-1] - y_c[-2]
        d1 = d1 / np.sqrt(d1**2 + d2**2)
        d2 = d2 / np.sqrt(d1**2 + d2**2)
        ax.arrow(x_c[-1], y_c[-1], d1, d2, width=0, lw=1, head_width=2,
                head_length=2, fc='k', ec='k')

        ax.grid(True)

    #label = io.filename2label(filename)
    #text = ', '.join([
    #    'Q = %.1f Mt/y' % x[0],
    #    'K = %d m$^2$/s' % x[1],
    #    'x$_{BG}$ = %d ppm' % x[2],
    #    'u = %.1f m/s' % wind
    #])
    #fig.suptitle('Gaussian plume fit for Janschwalde: %s\n(%s)' % (label, text), fontsize=10, y=0.99)


    for fig in [fig1, fig2]:
        fig.subplots_adjust(right=0.8)
        cax = fig.add_axes([0.82, 0.15, 0.03, 0.7])
        cb = fig.colorbar(c, cax=cax, extend='both')
        cb.set_label('XCO$_2$ (ppm)')


    # text: wind-speed


def estimate_plume_height(z, d, d0, h0=100):
    """\
    Estimate height of plume using wind direction of plume d0 and
    profile of wind direction d(z). The algorithm will minimize the difference
    between them.
    """
    def f(h, z, d, d0):
        d = np.interp(h,z,d)
        return misc.anglediff(d0,d)**2

    #TODO: does not work very well because of several local minima
    res = sopt.minimize(f, x0=h0, args=(z,d,d0))

    return res


def estimate_wind_speed_along_curve(xc, yc, x_model, y_model, height_model,
                                    uwind, vwind, f_curve, c_curve, x0, y0):
    heights = []
    speed = []
    direction = []
    direction_sat = []

    uu = []
    vv = []

    for x,y in zip(xc[::4],yc[::4]):
        i_o, j_o = misc.nearest_index(x, y, x_model, y_model)

        # estimate cardinal wind direction (and plume height) 
        # along center line (TODO)
        uwind2 = uwind[::-1, i_o, j_o][:30]
        vwind2 = vwind[::-1, i_o, j_o][:30]
        hh = height_model[::-1, i_o, j_o]
        hh = 0.5 * (hh[1:] + hh[:-1])[:30]

        heights.append(hh)
        uu.append(uwind2)
        vv.append(vwind2)

        u_mean = np.mean(uwind2[(200 <= hh) & (hh < 1000.0)])
        v_mean = np.mean(vwind2[(200 <= hh) & (hh < 1000.0)])

        speed.append( np.sqrt(uwind2**2 + vwind2**2))

        # wind direction (model)
        dd = (270.0 - np.rad2deg(np.arctan2(uwind2,vwind2))) % 360
        direction.append(dd)

        # wind direction (satellite)
        xt,yt = f_curve(c_curve, x=x, y=y, m=1, x0=x0, y0=y0)
        d_sat = (270 - np.rad2deg(np.arctan2(yt,xt))) % 360

        direction_sat.append(d_sat)

    return xc, yc, np.array(heights), np.array(speed), np.array(direction), \
            np.array(direction_sat), np.array(uu), np.array(vv)




def integrate_field(data, tracer, curve, xp, bg=None, tmax=100e3):

    # TODO: field -> molecules/m2
    # for NO2: times 1e4
    x = np.asarray(data.x)
    y = np.asarray(data.y)
    if isinstance(tracer, str):
        c = np.asarray(data[tracer])
    else:
        c = np.asarray(tracer)

    c[~data.plume_area] = np.nan

    # normal at xp
    yp = np.arange(-tmax, tmax + 0.1, 1e3)
    xn, yn = curve.compute_normal(t0=xp, t=yp)

    z = misc.interp_xy(xn, yn, x, y, c)

    if bg is None:
        bg = 0.0
    elif np.ndim(bg) > 1:
        bg = misc.interp_xy(xn,yn, x,y, bg)

    z_signal = z - bg
    valids = np.isfinite(z_signal)

    if np.any(valids):
        i = scipy.integrate.simps(z_signal[valids], yp[valids])
    else:
        i = np.nan


    return i






def compute_background(data, tracer, sigma=None, method='spline', show=False):
    """
    Interpolate smooth background.
    """
    x = np.asarray(data.x)
    y = np.asarray(data.y)

    if isinstance(tracer,str):
        c = np.asarray(data[tracer])
        sigma = data[tracer].noise_level
    else:
        c = np.array(tracer)

    area = np.asarray(data.plume_area)

    # only use pixels that are in plume area around plume without enhanced
    # tracer values
    valids = np.array(area & ~data.is_plume & ~data.is_hit)
    valids[~np.isfinite(c)] = False


    if method == 'spline':
        i = SmoothBivariateSpline(x[valids], y[valids], c[valids],
                                  w=np.full(valids.sum(), 1.0/sigma))

        smooth_field = np.full(c.shape, np.nan)
        smooth_field[area] = i.ev(x[area], y[area])
    else:
        sigma = 10.0 # in pixels
        size = int(5 * sigma)
        kernel = np.zeros((size,size))
        kernel[size//2,size//2] = 1.0
        kernel = ndi.gaussian_filter(kernel, sigma=sigma)

        c[~valids] = np.nan
        smooth_field = c2sm.misc.normalized_convolution(c, kernel, mask=~valids)
        smooth_field[~area] = np.nan

    if show:
        map.make_level2_map(data.lonc, data.latc, smooth_field,
                            vmin=405, vmax=405.5, do_zoom=True)


    return smooth_field


def plot_background():
    # plot background (true and estimated using different methods)
    pass

def main(data, origin, use_constant_emissions=False, model='2d',
         use_signal=False, use_noise=True, degree=2, offset=0,
        show=False, fig=None):

    # TODO
    # - 2D Gaussian plume fit
    # - 1D Gaussian plume fit with integration across center curve

    # read all data
    #data = io.read_level2(filename, 0.5, 0.01, 'high', 0.3)
    h,u,v = io.read_wind_profile(origin, data.time - timedelta(hours=offset))

    # setup flux inversion
    data, curve, params = setup_flux_inversion(
        data, origin=origin, height_model=h, uwind=u, vwind=v, degree=degree
    )


    if show:
        fig = plt.gcf() if fig is None else fig

        t = curve.t_o + np.linspace(0, data.xp.max())
        print(t)
        x,y = curve(t=t)
        map.add_curve(fig.axes[0], x,y, origin, label='center line')

    return data, curve, params

    dx, dy, field_vector, curve, wind = setup_flux_inversion(
        data, origin=origin, height_model=h, uwind=u, vwind=v, degree=degree
    )


    # do flux inversion

    print(wind)
    print('y shape:', field.shape)

    if model == '1d':
        x = np.nan
        res = {}

        # compute normals along plume center line

        # estimate XCO2 background and subtract (use BiSmoothingSpline
        # w=1/sigma)

        # integrate XCO2 signal across center lines

        # fit integrals to 1D gauss model

        # estimate uncertainties (instrument noise, wind, background...)


    elif model == '2d':
        x,res = flux_inversion([60.0,3000.0, 400], dx, dy, field_vector.a, wind, std=0.5)

        t = curve.get_parameter(field_vector.x, field_vector.y)
        t = np.linspace(t.min(), t.max())
        xc, yc = curve(t=t)

        if origin != 'Berlin':
            plume_age = 1e3 * t.max() / wind / 3600
            start = time - timedelta(hours=plume_age)
            true_emiss = io.read_emissions(origin, start, time,
                                           use_constant=use_constant_emissions)
            print('true emissions: (%.2f +/- %.2f) Mt/yr' % true_emiss)

        print('est. emissions: (%.2f +/- %.2f) Mt/yr' % (x[0], np.sqrt(res['S'][0,0])))

        visualize(xc,yc, field_vector, x, dx, dy, wind, plume_vector,
                  os.path.basename(filename))


    return x,res, curve, field, field_vector, dx, dy, clouds











