#!/usr/bin/env python
# coding: utf-8

from datetime import datetime, timedelta
import os

from matplotlib.dates import DateFormatter
import matplotlib.pyplot as plt
import numpy as np
import pandas
import scipy.ndimage as ndi
import scipy.optimize as opt

import doas

from smartcarb import io
from smartcarb import misc

import dplume

# kg/s -> Mt/y (Tg/y)
kgs_to_Mty = 1e3 * (365 * 24 * 60 * 60) / 1e12


def compute_emissions(nc):

    cats = [
       'CO2_point_snap1_major', # this was counted twice in Jan and Jul
       'CO2_point_snap1_minor',
       'CO2_point_snap2_minor',
       'CO2_point_snap3_minor',
       'CO2_point_snap4_minor',
       'CO2_point_snap5_minor',
       'CO2_point_snap34_minor',
       'CO2_point_snap8_minor',
       'CO2_point_snap9_minor',
       'CO2_area_construction',
       'CO2_line_aircraft',
       'CO2_point_aircraft',
       'CO2_area_aircraft',
       'CO2_area_forestry',
       'CO2_area_offroad',
       'CO2_line_shipping',
       'CO2_point_shipping',
       'CO2_line_train',
       'CO2_area_heating',
       'CO2_line_majorroads',
       'CO2_area_minorroads'
    ]

    total = 0.0
    for cat in cats:
        total += np.sum(nc.variables[cat][:] * 1235941.1)

    return total



def get_constant_emissions(time, use_constant=True):
    if not use_constant and time.year == 2015 and time.month in [1,7]:
        return 754.86
    else:
        return 534.30



def simple_leastsq(filename, tracer, thr, cloud_thr=0.01, error_std=0.7, seed='orbit',
        background_mode='model', use_constant_emissions=False, use_real_footprint=False
    ):
    """\
    filename is either a file with 2d COSMO field or Level-2 data

    y[i,j] = x * H[i,j] + b

    - y: XCO2 measurements in plume
    - H: XCO2 from Berlin tracer (e.g. XCO2_BC) normalized by emissions
    - b: background (e.g. mean of "background" tracers in COSMO)

    The plume is defined where 'plume_tracer' > 'plume_thr'.

    If `use_const_emissions` use XCO2_BV in observation vector y.
    """
    # tracer fields (synthetic observations)
    time = io.filename2time(filename)

    # define plume based on threshold (use BV or BC)
    p = dplume.GroundTruth(filename, 'Berlin', tracer='CO2', thr=0.05, cloud_thr=0.3)
    is_plume = p.plume & (p.signal > thr) & (p.clouds < cloud_thr)


    bv, bc = io.read_fields(filename, ['XCO2_BV', 'XCO2_BC'], time, correct_berlin=True)

    co2 = io.read_total_tracer(filename, 'CO2', time, correct_berlin=True,
            use_constant_emissions=use_constant_emissions)

    noise = io.read_tracer_noise(filename, tracer, True, error_std, seed=seed)



    if np.sum(is_plume) <= 1:
        return np.nan, np.nan, np.sum(is_plume)


    # background
    if use_constant_emissions:
        b = (co2 - bc)[is_plume]
    else:
        b = (co2 - bv)[is_plume]


    if background_mode == 'model':
        bg_std = 0.0
    elif background_mode == 'mean':
        bg_std = b.std()
        b = b.mean()
    else:
        raise ValueError('unknown mode for background estimation')

    # observation vector
    y = co2[is_plume]

    # footprint
    if use_real_footprint and not use_constant_emissions:
        H = bv[is_plume] / 534.30
    else:
        H = bc[is_plume] / 534.30

    # add noise
    y += noise[is_plume]

    # Se, xa and Sa
    Se = np.nan

    # forward model
    f = lambda x,_: H * x / kgs_to_Mty + b

    x0 = np.array([534.0]) * kgs_to_Mty
    x, res = doas.solver.gauss_newton(x0, y, f, Se=Se)

    return x[0], np.sqrt(float(res['S'])), np.sum(is_plume)



def main(out_filename, thr, error_std=0.7, nmin=100,
         use_constant_emissions=False, background_mode='model'):

    filename = '/home/kug/projects/smartcarb/data/Berlin_Overpasses6.dat'
    data = pandas.read_csv(filename, index_col=0, parse_dates=True)
    level2_filenames = [
        io.create_filename(data.index[i], None, 'level2', 'Sentinel_7_CO2',
                           data.orbits[i], data.lon_eq[i])
        for i in range(data.shape[0])
        if data['cloud-free pixels'][i] > nmin
    ]

    em = pandas.read_csv('/project/smartcarb/flux_inversion/ts_berlin_emissions.dat',
                         delimiter='\t', index_col=0, parse_dates=True)
    em = em['CO2_BV_fix']

    with open(out_filename, 'w') as out_file:
        out_file.write('\t'.join([
            'UTC',
            'lon_eq',
            'orbit',
            'number of pixels',
            'CO2 emissions (Mt/yr)',
            'CO2 uncertainty (Mt/yr)',
            'CO2 true emissions (Mt/yr)',
        ]))
        out_file.write('\n')

        for fn in level2_filenames:

            _, orbit, time, lon_eq = io.process_filename(fn)

            x, x_std, N = simple_leastsq(
                fn, 'CO2', thr, error_std=error_std,
                use_constant_emissions=use_constant_emissions,
                background_mode=background_mode
            )

            if use_constant_emissions:
                true_em = 16.849540
            else:
                true_em = em[time]

            print(time, x, true_em)

            out_file.write('\t'.join([
                str(time),
                str(lon_eq),
                str(orbit),
                str(N),
                '%10.3f' % x,
                '%10.3f' % x_std,
                '%10.3f' % true_em
            ]))
            out_file.write('\n')





def plot_ts(filename, nsat=1):


    ts = pandas.read_csv(filename, delimiter='\t', index_col=0,
                         parse_dates=True, na_values='%10.3f' % np.nan)

    # filter constellation
    DLON_ORBIT = 24.146341463414636
    lon_eqs = np.linspace(0, DLON_ORBIT, nsat+1)[:-1]
    sats = ts.lon_eq == 0

    for lon_eq in lon_eqs[1:]:
        lon_eq = int(100 * round(lon_eq, 2))
        sats |= np.abs(ts.lon_eq - lon_eq) < 1


    sats &= 0.0 <= ts['CO2 emissions (Mt/yr)']
    sats &= ts['CO2 emissions (Mt/yr)'] <= 40

    ts = ts[sats]


    est = ts['CO2 emissions (Mt/yr)']
    true = ts['CO2 true emissions (Mt/yr)']

    MB = np.mean(est - true)
    NMB = 100 * np.mean( (est - true) / true)
    SD = np.std(est - true)



    print('MB = %.2f (%d), SD = %.2f (%d)' % (MB, NMB, SD, 100 * SD / true.mean()))


    em = pandas.read_csv('/project/smartcarb/flux_inversion/ts_berlin_emissions.dat',
                         delimiter='\t', index_col=0, parse_dates=True)
    em = em['CO2_BV_fix']


    fig = plt.figure()
    ax = fig.add_subplot(111)

    mask = em.index.hour == 11
    ax.plot(em[mask], label='true emissions (11 UTC)')

    ax.errorbar(ts.index, ts['CO2 emissions (Mt/yr)'],
                ts['CO2 uncertainty (Mt/yr)'], ls='', marker='o',
                label='estimated emissions')

    ax.xaxis.set_major_formatter(DateFormatter('%b'))

    ax.set_xlim('2015-01-01', '2015-12-31')
    ax.set_ylim(0,50)

    ax.set_xlabel('2015')
    ax.set_ylabel('CO$_2$ emissions (Mt/yr)')
    ax.grid(True)


    plt.legend(loc=1)



    return est, true



